/*
 *    bookkeeping.C --- Contains simple, short function without mathematical
 *                      ideas
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <list>
#include <vector>
#include <deque>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"

void one::eraseWeb(unsigned int idx) {
    webs.at(idx)->deleteContents();
    delete webs.at(idx);
    webs.at(idx) = 0;
}

void chainComplex::getMinMaxQ(int &minQ, int &maxQ) {
    bool notset = true;
    for (std::vector<one*>::iterator voIt = spaces.begin();
            voIt != spaces.end(); ++voIt)
        for (WEBCONT<web*>::iterator lwIt = (*voIt)->webs.begin();
                lwIt != (*voIt)->webs.end(); ++lwIt) {
            if (!*lwIt) continue;
            int q = (*lwIt)->qShift;
            if (notset || (q < minQ)) minQ = q;
            if (notset || (q > maxQ)) maxQ = q;
            notset = false;
        }
}

void facet::eraseEdge(const iEdge &e) {
    for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator i = boundaries.begin();
            i != boundaries.end(); ++i)
        for (EDGECONT<edge*>::iterator j = i->begin(); j != i->end(); ++j)
            if (*j == &e) {
                i->erase(j);
                if (i->empty())
                    boundaries.erase(i);
                return;
            }
}

ratio lcfoams::sumOfCoefficients() const {
    ratio result = ratio(0);
    for (FACETCONT<foam*>::const_iterator it = foams.begin();
            it != foams.end(); ++it)
        result += (*it)->coefficient;
    return result;
}

bool foam::isInt() const {
    return (facets.empty());
}

bool lcfoams::isInt() const {
    for (FACETCONT<foam*>::const_iterator i = foams.begin();
            i != foams.end(); ++i)
        if (! (*i)->isInt()) return false;
    return true;
}

oPoint::oPoint(bool increaseCounter) : hasAlreadyOccurred(false),
    isGlued(false) {
    isGlued = false;
    isInner = false;
    if (increaseCounter) {
        number = oPointMax++;
        oPointIdx.push_back(this);
    }
}

oEdge::oEdge() : idFacet(0), isBelow(false) {
    points[0] = points[1] = 0; isInner = false;
}

iPoint::iPoint() : idSingular(0) {
    copyTag = 0;
    edges[0] = edges[1] = edges[2] = 0;
    isInner = true;
}

void iPoint::saveEdges() {
    for (int i = 0; i < 3; ++i)
        oldEdges[i] = edges[i];
}

void iPoint::restoreEdges() {
    for (int i = 0; i < 3; ++i)
        edges[i] = oldEdges[i];
}

bool iPoint::isTransit() const {
    return (edges[0] && !edges[2]);
}

bool iPoint::isIncoming() const {
    return (edges[0] && (edges[0]->points[1] == this));
}

oEdge* iPoint::otherEdge(const oEdge &oneEdge) const {
    return (&oneEdge) == edges[0] ? edges[1] : edges[0];
}
        
void chainComplex::deleteContents() {
    for (unsigned i = 0; i < maps.size(); ++i) {
        maps.at(i)->deleteContents();
        delete maps.at(i);
    }
    for (unsigned i = 0; i < spaces.size(); ++i) {
        spaces.at(i)->deleteContents();
        delete spaces.at(i);
    }
}

void web::setEdgesToBelow() const {
    for (OEDGECONT<oEdge*>::const_iterator i = edges.begin();
            i != edges.end(); ++i)
        (*i)->isBelow = true;
}

void web::setEdgesToAbove() const {
    for (OEDGECONT<oEdge*>::const_iterator i = edges.begin();
            i != edges.end(); ++i)
        (*i)->isBelow = false;
}

void web::deleteContents() {
    for (IPOINTCONT<iPoint*>::iterator i = points.begin();
            i != points.end(); ++i)
        delete *i;
    for (OEDGECONT<oEdge*>::iterator i = edges.begin(); i != edges.end(); ++i)
        delete *i;
}

void foam::deleteContents() {
    for (FACETCONT<facet*>::iterator i = facets.begin(); i != facets.end(); ++i)
        delete *i;
    for (SINGULARCONT<iEdge*>::iterator i = singulars.begin();
            i != singulars.end(); ++i)
        delete *i;
}

void lcfoams::deleteContents() {
    for (FOAMCONT<foam*>::iterator i = foams.begin(); i != foams.end(); ++i) {
        (*i)->deleteContents();
        delete *i;
    }
}

bool lcfoams::isZero() const {
    return (foams.size() == 0) ||
        (isInt() && (! sumOfCoefficients().isNonZero()));
}

void lcfoams::composeLcFoams(foam *other, const web & below,
        const web & below2, bool deleteIt) {
    for (FOAMCONT<foam*>::iterator i = foams.begin(); i != foams.end(); ++i)
        (*i)->composeFoam(other->copy(), below, below2);
    if (deleteIt) {
        other->deleteContents();
        delete other;
    }
}

void lcfoams::glueLcfoams(foam *other, bool dualizeIt, bool deleteIt) {
    for (FOAMCONT<foam*>::iterator i = foams.begin(); i != foams.end(); ++i)
        (*i)->glueFoam(other->copy(dualizeIt));
    if (deleteIt) {
        other->deleteContents();
        delete other;
    }
}

void lcfoams::changeSign() {
    for (FOAMCONT<foam*>::iterator i = foams.begin();
            i != foams.end(); ++i) (*i)->changeSign();
}

void lcfoams::dualize() {
    for (FOAMCONT<foam*>::iterator i = foams.begin();
            i != foams.end(); ++i) (*i)->dualize();
}

lcfoams* lcfoams::copyNoDualAndDeleteZero() {
    lcfoams* result = copy();
    if (result->foams.empty()) {
        delete result;
        result = 0;
    }
    return result;
}

lcfoams::lcfoams(foam* aSingleFoam) {
    doGlueSS = false;
    foams.push_back(aSingleFoam);
}

unsigned int one::realSize() const {
    unsigned int result = 0;
    for (WEBCONT<web*>::const_iterator i = webs.begin(); i != webs.end(); ++i)
        if (*i) ++result;
    return result;
}

void one::deleteContents() {
    for (WEBCONT<web*>::iterator i = webs.begin(); i != webs.end(); ++i) {
        if (*i) {
            (*i)->deleteContents();
            delete *i;
        }
    }
}

void two::deleteContents() {
    for (MatrixIterator<lcfoams> i = this->begin();
            ! i.isFinished(); i.step()) {
        i->deleteContents();
        delete *i;
    }
    this->resize(0, 0);
}

void two::checkForZeroes() {
    MatrixIterator<lcfoams> mei = this->begin();
    while (! mei.isFinished()) {
        if (mei->foams.empty()) {
            mei->deleteContents();
            delete *mei;
            erase(&mei);
        } else mei.step();
    }
}

two::two(foam *f) { 
    if (f) {
        this->resize(1, 1);
        set(0, 0, new lcfoams(f));
    }
}

void chainComplex::setToSmall() {
    spaces.push_back(new one());
    (*spaces.begin())->webs.push_back(new web());
}
