/*
 *    dtToPd.C --- Convert Dokwer-Thistlethwaite to Planar-Diagram code.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <vector>
#include <stdlib.h>
#include <iostream>

#include "dtToPd.H"

UnsignedIntList* emptyUnsignedIntList();
UnsignedIntList* newUnsignedIntList(unsigned newData);
UnsignedIntList *addToUnsignedIntList(UnsignedIntList **l, unsigned x);
UnsignedIntList *insertAfterUnsignedIntList(UnsignedIntList *l,
        unsigned value);
bool signIt(UnsignedIntList *position, unsigned movingIn, unsigned const *order,
        unsigned segment, std::vector<bool> const &input_signs);
bool secondPart(UnsignedIntList *position, unsigned *region, unsigned segment,
        unsigned currentRegion, unsigned newRegion, unsigned const *order,
        std::vector<bool> const &input_signs, std::vector<bool> &output_signs,
        unsigned nbCrossings, unsigned *c, unsigned const *linkComponents,
        bool changeRegion);
bool dtSigns(UnsignedIntList *position, unsigned *region, unsigned segment,
        unsigned movingIn, unsigned movingOut, unsigned currentRegion,
        unsigned newRegion, unsigned const *order,
        std::vector<bool> const &input_signs, std::vector<bool> &output_signs,
        unsigned nbCrossings, unsigned *c, unsigned const *linkComponents,
        bool changeRegion);

UnsignedIntList* emptyUnsignedIntList() {
    UnsignedIntList* newL;
    newL = (UnsignedIntList*) malloc(sizeof(UnsignedIntList));
    if (!newL) {
        std::cerr << "Error allocating memory for new UnsignedIntList.\n";
        exit(EXIT_FAILURE);
    }
    newL->next = 0;
    return newL;
}

UnsignedIntList* newUnsignedIntList(unsigned newData) {
    UnsignedIntList* newL = emptyUnsignedIntList();
    newL->data = newData;
    return newL;
}

UnsignedIntList *addToUnsignedIntList(UnsignedIntList **l, unsigned x) {
	UnsignedIntList *newL, **end = l;
	while (*end) end = &((*end)->next);
	newL = newUnsignedIntList(x);
	return *end = newL;
}

UnsignedIntList *insertAfterUnsignedIntList(UnsignedIntList *l,
        unsigned value) {
    UnsignedIntList *newL = newUnsignedIntList(value); 
    newL->next = l->next;
    return (l->next = newL);
}

bool signIt(UnsignedIntList *position, unsigned movingIn, unsigned const *order,
        unsigned segment, std::vector<bool> const &input_signs) {
    UnsignedIntList *k = position;
    while (k->next->data != movingIn)
        k = k->next;

    bool result = (k->next->data + segment) % 2;
    if (input_signs.at(order[segment - 1])) result = ! result;
    return result;
}

bool secondPart(UnsignedIntList *position, unsigned *region, unsigned segment,
        unsigned currentRegion, unsigned newRegion, unsigned const *order,
        std::vector<bool> const &input_signs, std::vector<bool> &output_signs,
        unsigned nbCrossings, unsigned *c, unsigned const *linkComponents,
        bool changeRegion) {
    if (!region[segment * 2]) {
        /* A new end
         * Insert two entries between position and position->next
         */
        position = insertAfterUnsignedIntList(position, 2 * c[segment]);
        insertAfterUnsignedIntList(position, 2 * c[segment] + 1);
        region[c[segment] * 2] = region[c[segment] * 2 + 1] =
            region[2 * segment] = region[2 * segment + 1] = currentRegion;
        return dtSigns(position, region, segment + 1, -1, -1,
                    currentRegion, newRegion, order, input_signs, output_signs,
                    nbCrossings, c, linkComponents, changeRegion);
    } else {
        /* We already had that segment */
        int plus0 = (region[2 * segment] == currentRegion);
        int plus1 = (region[2 * segment + 1] == currentRegion);
        if (plus0 && plus1) {
            /* Two recursive calls
             * Make copies of position and region.
             */
            unsigned oldRegion[4 * nbCrossings];
            for (unsigned i = 0; i < 4 * nbCrossings; ++i)
                oldRegion[i] = region[i];
            UnsignedIntList *oldPosition = 0;
            {UnsignedIntList *j = position, *prev;
            do {
                prev = addToUnsignedIntList(&oldPosition, j->data);
                j = j->next;
            } while (j != position);
            prev->next = oldPosition;} /* make it circular */

            unsigned OneSeg = 2 * segment;
            unsigned OtherSeg = 2 * segment + 1;

            UnsignedIntList *oldPosition2 = 0;
            {UnsignedIntList *j = position, *prev;
            do {
                prev = addToUnsignedIntList(&oldPosition2, j->data);
                j = j->next;
            } while (j != position);
            prev->next = oldPosition2;} /* make it circular */
            if (signIt(oldPosition2, OneSeg, order, segment + 1, input_signs)
                    != input_signs.at(order[segment])) {
                ++OneSeg;
                --OtherSeg;
            }
            if (dtSigns(position, region, segment + 1, OneSeg, OtherSeg,
                    currentRegion, newRegion, order, input_signs, output_signs,
                    nbCrossings, c, linkComponents, changeRegion))
                return true;
            else return dtSigns(oldPosition, oldRegion, segment + 1,
                    OtherSeg, OneSeg, currentRegion, newRegion, order,
                    input_signs, output_signs, nbCrossings, c, linkComponents,
                    changeRegion);
        } else if (plus0 || plus1) {
            /* One recursive call */
            return dtSigns(position, region, segment + 1,
                    plus0 ? 2 * segment : 2 * segment + 1,
                    plus0 ? 2 * segment + 1 : 2 * segment,
                    currentRegion, newRegion, order, input_signs, output_signs,
                    nbCrossings, c, linkComponents, changeRegion);
        } else return false;
    }
}

bool dtSigns(UnsignedIntList *position, unsigned *region, unsigned segment,
        unsigned movingIn, unsigned movingOut, unsigned currentRegion,
        unsigned newRegion, unsigned const *order,
        std::vector<bool> const &input_signs, std::vector<bool> &output_signs,
        unsigned nbCrossings, unsigned *c, unsigned const *linkComponents,
        bool changeRegion) {
    if (movingIn != (unsigned)(-1)) {
        output_signs.at(order[segment - 1]) = signIt(position, movingIn, order,
                segment, input_signs);

        UnsignedIntList *k = position;
        while (k->next->data != movingIn) {
            if (changeRegion && (region[k->next->data] == currentRegion))
                region[k->next->data] = newRegion;
            k = k->next;
        }

        currentRegion = region[movingOut];
        newRegion++;

        free(k->next);
        k->next = k->next->next;
        while (k->next->data != movingOut) {
            k = k->next;
        } 
        free(k->next);
        k->next = k->next->next;
        position = k;
    }
    if (segment == nbCrossings * 2) return true;
    /* Now change segment if a new link component begins */
    if (segment != linkComponents[segment]) {
        /* Recursively try all regions */
        for (unsigned rIter = 1; rIter < newRegion; ++rIter) {
            /* save */
            unsigned oldRegion[4 * nbCrossings];
            for (unsigned i = 0; i < 4 * nbCrossings; ++i)
                oldRegion[i] = region[i];
            UnsignedIntList *oldPosition = 0;
            {UnsignedIntList *j = position, *prev;
            do {
                prev = addToUnsignedIntList(&oldPosition, j->data);
                j = j->next;
            } while (j != position);
            prev->next = oldPosition;}

            if (secondPart(position, region, segment,
                        rIter, newRegion, order, input_signs, output_signs,
                        nbCrossings, c, linkComponents, false)) return true;

            /* restore */
            for (unsigned i = 0; i < 4 * nbCrossings; ++i)
                region[i] = oldRegion[i];
            position = oldPosition;
        }
        return false;
    }
    return secondPart(position, region, segment,
            currentRegion, newRegion, order, input_signs, output_signs,
            nbCrossings, c, linkComponents, true);
}
