/*
 *    main.C --- Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <algorithm>
#include <time.h>
#include <iostream>
#include <iomanip>
#include <list>
#include <vector>
#include <deque>

#include "getMem.H"
#include "macros.H"
#include "optionArray.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"
#include "smith.H"
#include "dtToPd.H"
#include "parser.H"
#include "output.H"

const unsigned GreedyIterations = 1000;
bool ratio::useRationals;
std::vector<oPoint*> oPoint::oPointIdx;
unsigned oPoint::oPointMax = 0;

void displayFloat(float f);
bool isIncoming(unsigned idx, unsigned type);
void simplifyIt(chainComplex &c, int verbosityLevel, bool markedReduced);
int minimizeWidthGreedily(VECVECINT &v);
unsigned computeRasmussen(unsigned ccSpacesSize, int tShift, unsigned QIdx,
        std::vector<std::vector<unsigned> > homDims);
void mainAlgorithm(int argc, char* argv[], time_t startTime,
        long unsigned &MemoryConsumption, bool &displayTimeMemory);

void displayFloat(float f) {
    std::cout << (int)(f) << "." << abs((int)(10 * f) - 10*(int)(f));
}

bool isIncoming(unsigned idx, unsigned type) {
    switch (type) {
        case POS_CROS : return ((idx == 0) || (idx == 3));
        case NEG_CROS : return ((idx == 0) || (idx == 1));
        case SEGMENT : return (idx == 0);
        default: throw 32;
    }
}

void simplifyIt(chainComplex &c, int verbosityLevel, bool markedReduced) {
    if (verbosityLevel > 0) std::cout << "Simplifying:\n";
    for (unsigned i = 0; i < c.spaces.size(); ++i) {
        std::deque<std::vector<unsigned> > gaussStack;
        int dids = 0;
        if (verbosityLevel > 1)
            std::cout << MARK_UP "At t-index " << i << NO_MARK_UP ": "
                << std::flush;
        else if (verbosityLevel == 1) {
            if (i == 0)
                std::cout << "At t-index " << std::flush;
            std::cout << i << "; " << std::flush;
        }
        int counter = 0;
        c.spaces.at(i)->oneSimplify(i > 0 ? c.maps.at(i - 1) : 0,
                i < c.spaces.size() - 1 ? c.maps.at(i) : 0,
                true, 0, i - 1, markedReduced);
        while  (dids != -1) {
            dids = c.spaces.at(i)->oneSimplify(i > 0 ? c.maps.at(i - 1) : 0,
                    i < c.spaces.size() - 1 ? c.maps.at(i) : 0,
                    false, &gaussStack, i - 1, markedReduced);
            if (verbosityLevel && (dids != -1)) {
                ++counter;
                if (verbosityLevel > 1)
                    std::cout << (2 * dids - 2) << std::flush;
            }
            while (! gaussStack.empty()) {
                std::vector<unsigned> para = gaussStack.back();
                gaussStack.pop_back();
                if (c.gauss(para.at(0), para.at(1), para.at(2),
                            &gaussStack, true) && verbosityLevel > 1)
                    std::cout << "~";
            }
        }
        if (verbosityLevel > 1) std::cout << "\n";
    }   
    if (verbosityLevel) std::cout << "\n";
}

int minimizeWidthGreedily(VECVECINT &v) {
    int bestWidth = -1;
    for (unsigned randIter = 0; randIter < GreedyIterations; ++randIter) {
        VECVECINT w;
        std::vector<bool> done;
        done.resize(v.size(), false);
        w.reserve(v.size());
        std::vector<bool> isIn;
        isIn.resize(2 * v.size(), false);
        int width = 0;
        int maxWidth = 0;
        for (unsigned i = 0; i < v.size(); ++i) {
            unsigned maxIdx = 0;
            int maxOld = -1;
            for (unsigned j = 0; j < v.size(); ++j) {
                if (! done.at(j)) {
                    int old = 0;
                    for (unsigned k = 0; k < v.at(j).size() - 1; ++k)
                        if (isIn.at(v.at(j).at(k))) ++old;
                    if (old > maxOld) {
                        maxOld = old;
                        maxIdx = j;
                    }
                }
            }
            for (unsigned k = 0; k < v.at(maxIdx).size() - 1; ++k) {
                width += isIn.at(v.at(maxIdx).at(k)) ? -1 : 1;
                isIn.at(v.at(maxIdx).at(k)) = true;
            }
            maxWidth = (maxWidth < width) ? width : maxWidth;
            w.push_back(v.at(maxIdx));
            done.at(maxIdx) = true;
        }
        if ((bestWidth == -1) || (maxWidth < bestWidth)) {
            std::random_shuffle(v.begin(), v.end());
            v.swap(w);
            bestWidth = maxWidth;
        }
    }
    return bestWidth;
}

void chainComplex::computeHomology(bool rational, unsigned &QIdx,
        int &minQ, std::vector<std::vector<unsigned> > &homDims,
    std::vector<torstruct> &torsion) {
    {
        int maxQ;
        getMinMaxQ(minQ, maxQ);
        QIdx = (maxQ - minQ + 2) / 2;
    }

    std::vector<std::vector<unsigned> > ccDims;
    ccDims.resize(spaces.size());

    std::vector<std::vector<unsigned> > IdxTranslator;
    IdxTranslator.resize(spaces.size());
    for (unsigned i = 0; i < spaces.size(); ++i) {
        IdxTranslator.at(i).resize(spaces.at(i)->webs.size(),
                (unsigned)(-1));
        ccDims.at(i).resize(QIdx);
        unsigned j = 0;
        for (WEBCONT<web*>::iterator lwIt = spaces.at(i)->webs.begin();
                lwIt != spaces.at(i)->webs.end(); ++lwIt, ++j) {
            if (!*lwIt) continue;
            const int q = (*lwIt)->qShift;
            IdxTranslator.at(i).at(j) = ccDims.at(i).at((q - minQ) / 2);
            ccDims.at(i).at((q - minQ) / 2) += 1;
        }
    }

    std::vector<std::vector<unsigned> > ranks;
    ranks.resize(maps.size());
    for (unsigned i = 0; i < maps.size(); ++i) ranks.at(i).resize(QIdx);
    if (! rational)
        smithNormalForm(this, torsion, ranks, ccDims, IdxTranslator,
                spaces.size(), QIdx, minQ);

    homDims.resize(spaces.size());
    for (unsigned i = 0; i < spaces.size(); ++i) {
        homDims.at(i).resize(QIdx);
        for (unsigned j = 0; j < QIdx; ++j)
            homDims.at(i).at(j) = ccDims.at(i).at(j) -
                ((i < spaces.size() - 1) ? ranks.at(i).at(j) : 0) - 
                ((i > 0) ? ranks.at(i-1).at(j) : 0);
    }
}

unsigned computeRasmussen(unsigned ccSpacesSize, int tShift, unsigned QIdx,
        std::vector<std::vector<unsigned> > homDims) {
    for (int i = 0; i < - tShift; ++i) {
        if (homDims.at(i).at(0) || homDims.at(i).at(1) || homDims.at(i).at(2))
            return false;
        for (unsigned j = 3; j < QIdx; ++j) {
            if (homDims.at(i).at(j) > homDims.at(i + 1).at(j - 3))
                return NO_RASMUSSEN;
            else {
                homDims.at(i + 1).at(j - 3) -= homDims.at(i).at(j);
                homDims.at(i).at(j) = 0;
            }
        }
    }
    for (int i = ccSpacesSize - 1; i > - tShift; --i) {
        if (homDims.at(i).at(QIdx - 1) || homDims.at(i).at(QIdx - 2) ||
                homDims.at(i).at(QIdx - 3))
            return false;
        for (unsigned j = 0; j < QIdx - 3; ++j) {
            if (homDims.at(i).at(j) > homDims.at(i - 1).at(j + 3))
                return NO_RASMUSSEN;
            else {
                homDims.at(i - 1).at(j + 3) -= homDims.at(i).at(j);
                homDims.at(i).at(j) = 0;
            }
        }
    }
    unsigned rasmussen = (unsigned)(-1);
    for (unsigned j = 0; j < QIdx; ++j) 
        if (homDims.at(-tShift).at(j)) {
            rasmussen = j;
            break;
        }
    if (rasmussen == (unsigned)(-1)) return NO_RASMUSSEN;
    for (unsigned j = 0; j < QIdx; ++j)  {
        unsigned expectedValue = ((j == rasmussen) || (j == rasmussen + 1) ||
                (j == rasmussen + 2)) ? 1 : 0;
        if (homDims.at(-tShift).at(j) != expectedValue)
            return NO_RASMUSSEN;
    }
    return 2 * rasmussen + 2;
}

void mainAlgorithm(int argc, char* argv[], time_t startTime,
        long unsigned &MemoryConsumption, bool &displayTimeMemory) {
    optionArray oa;
    switch (parseCommandLine(argc, argv, oa)) {
        case OPTION_TITLE: printTitle(); return;
        case OPTION_HELP:  printHelp(); return;
        case OPTION_ERROR: return;
        default: ;
    }
    if ((! withPari()) && (! oa.rational)) {
        std::cerr << "foamho has been compiled without the PARI-library, so the"
            " smith normal form, and hence the torsion subgroup of the homology"
            " cannot be calculated. Computing rational homology.\n";
        oa.rational = true;
    }

    ratio::useRationals = oa.rational;
    displayTimeMemory = oa.displayTimeMemory;

    chainComplex c;
    c.setToSmall();

    if (oa.optimiseGirth) {
        int girth = minimizeWidthGreedily(oa.v);
        std::cout << "Girth-optimised link diagram (modified pd notation) [";
        printVVi(oa.v);
        std::cout << "Girth: " << girth << ".\n";
    } else {
        std::cout << "Working on link diagram (modified pd notation) [";
        printVVi(oa.v);
    }
    if (oa.reduced) {
        int maxSeg = 0;
        for (VECVECINT::iterator vviIt = oa.v.begin();
                vviIt != oa.v.end(); ++vviIt)
            for (unsigned l = 0; l < vviIt->size() - 1; ++l)
                if (vviIt->at(l) >= maxSeg)
                    maxSeg = vviIt->at(l) + 1;
        if (oa.redStrand && ((int)(oa.redStrand) > maxSeg)) {
            std::cerr << "The segment number must be between 1 and " << maxSeg << ".\n";
            return;
        }
        int insertSegmentAt = oa.redStrand ? oa.redStrand - 1 :
            oa.v.back().front();
        for (VECVECINT::iterator vviIt = oa.v.begin();
                vviIt != oa.v.end(); ++vviIt)
            for (unsigned l = 0; l < vviIt->size() - 1; ++l)
                if ((vviIt->at(l) > insertSegmentAt) ||
                        ((vviIt->at(l) == insertSegmentAt) &&
                         isIncoming(l, vviIt->back()))) 
                    vviIt->at(l) += 1;
        VECINT pointedSeg;
        pointedSeg.push_back(insertSegmentAt);
        pointedSeg.push_back(insertSegmentAt + 1);
        pointedSeg.push_back(POINTED_SEGMENT);
        oa.v.push_back(pointedSeg);
    }
    unsigned idx = 0;
    int k = 0;
    int girth = 0;
    std::cout << "Calculating...\n";
    for (VECVECINT::iterator vviIt = oa.v.begin();
            vviIt != oa.v.end(); ++vviIt) {
        {
            unsigned max = 0;
            for (VECINT::iterator viIt = vviIt->begin();
                    viIt != --(vviIt->end()); ++viIt) {
                if (*viIt < 0) throw 31;
                if ((unsigned)(*viIt) > max) max = *viIt;
            }
            if (max >= oPoint::oPointIdx.size())
                for (int i = max - oPoint::oPointIdx.size() + 1; i > 0; --i)
                    new oPoint(true);
        }
        for (unsigned i = 0; i < vviIt->size() - 1; ++i) {
            girth += (oPoint::oPointIdx.at(vviIt->at(i))->hasAlreadyOccurred) ?
                -1 : 1;
            oPoint::oPointIdx.at(vviIt->at(i))->hasAlreadyOccurred = true;
        }
        if (oa.verbosityLevel > 0) {
            std::cout << "Tensoring with subtangle number " << ++k <<
                ". Girth is " << girth << ". ";
            if (oa.displayTimeMemory) {
                std::cout << "Runtime " << time(0) - startTime <<
                    " seconds Memory consumption: ";
                displayFloat((float)(getMemoryConsumption()) / 1000000);
                std::cout << " MB. ";
            }
            std::cout << std::flush;
        }
        const bool isLast = (vviIt + 1 == oa.v.end());

        const int action = vviIt->back();
        switch (action) {
            case POS_CROS: c.tensorPosCros(oPoint::oPointIdx.at(vviIt->at(0)),
                                   oPoint::oPointIdx.at(vviIt->at(1)),
                                   oPoint::oPointIdx.at(vviIt->at(2)),
                                   oPoint::oPointIdx.at(vviIt->at(3))); break;
            case NEG_CROS: c.tensorNegCros(oPoint::oPointIdx.at(vviIt->at(0)),
                                   oPoint::oPointIdx.at(vviIt->at(1)),
                                   oPoint::oPointIdx.at(vviIt->at(2)),
                                   oPoint::oPointIdx.at(vviIt->at(3))); break;
            case POINTED_SEGMENT:
            case SEGMENT: c.tensorLine(oPoint::oPointIdx.at(vviIt->at(0)),
                                  oPoint::oPointIdx.at(vviIt->at(1)));
                          break;
            default: std::cerr << "Unknown action: " << action << ".\n"; return;
        }
        simplifyIt(c, oa.verbosityLevel, oa.reduced && isLast);
        ++idx;
    }

    MemoryConsumption = getMemoryConsumption();

    for (unsigned i = 0; i < oPoint::oPointIdx.size(); ++i)
        delete oPoint::oPointIdx.at(i);
    oPoint::oPointIdx.clear();

    unsigned QIdx;
    int minQ;
    std::vector<std::vector<unsigned> > homDims;
    std::vector<torstruct> torsion;
    c.computeHomology(oa.rational, QIdx, minQ, homDims, torsion);

    std::cout << "Done. Result:\n";
    printHomology(c.spaces.size(), c.tShift, QIdx, minQ + c.qShift, homDims,
            torsion, oa.rational, oa.reduced);
    if (! oa.reduced)
        printRasmussen(
                computeRasmussen(c.spaces.size(), c.tShift, QIdx, homDims),
                minQ + c.qShift);
    c.deleteContents();
}

int main(int argc, char* argv[]) {
    try {
        time_t startTime = time(0);

        long unsigned MemoryConsumption = 0;
        bool displayTimeMemory = false;

        mainAlgorithm(argc, argv, startTime, MemoryConsumption,
                displayTimeMemory);

        if (displayTimeMemory) {
            std::cout << "Run time in seconds: " << time(0) - startTime << "\n";
            std::cout << "Memory consumption in megabytes: ";
            displayFloat((float)(MemoryConsumption) / 1000000);
            std::cout << "\n";
        }

        return EXIT_SUCCESS;
    } catch (int e) {
        std::cerr << "\nError " << e << " occurred. This is probably a bug. "
            "Please report the input and the output to "
            "lewark@math.jussieu.fr.\n";
    } catch (std::bad_alloc& ba) {
        std::cerr << "\nError: Ran out of memory when allocating (" <<
            ba.what() << "). Probably your link is just too big for the "
            "memory.\n";
    }
    return EXIT_FAILURE;
}
