/*
 *    main.H --- Foamho, a sl3-homology calculator.
 *               This file contains the definitions of most of the classes.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

/* point   = oPoint or iPoint
 * oPoint  = a boundary point of a web, 
 * edge    = oEdge or Iedge
 * oEdge   = an edge of a web, possibly a circle
 * iPoint  = a vertex of a web
 * facet   = a facet of a foam
 * iEdge   = a singular edge in a foam, possibly a circle
 * web     = a web
 * foam    = a foam in a linear combination of foams
 * lcfoams = a Z-linear combination of foams
 * one     = a vector of q-shifted webs
 * two     = a sparse matrix of lcfoams (heir of class SparseMatrix,
 *              see sparmat.H
 */
class point;
class oPoint;
class edge;
class oEdge;
class iPoint;
class facet;
class iEdge;
class web;
class foam;
class lcfoams;
class one;
class two;
class chainComplex;

/* to store integral homology as computed by smithNormalForm (see smith.H) */
typedef struct {
    std::vector<std::vector<unsigned> > dim;
    unsigned mod;
} torstruct;

class point {
    public:
        bool isInner;

        oPoint* asOPoint() { return (oPoint*)(this); }
        iPoint* asIPoint() { return (iPoint*)(this); }
};

class oPoint : public point {
    public:
        /* all oPoints are stored in this vector */
        static std::vector<oPoint*> oPointIdx;
        static unsigned oPointMax;

        /* index in oPointIdx */
        unsigned number;
        /* temporary, used only for girth minimising */
        bool hasAlreadyOccurred;

        /* temporary, used for gluing and composition */
        bool isGlued;

        oPoint(bool increaseCounter = false);
};

class edge {
    public:
        bool isInner;

        /* temporary, points to copy of this when copying a whole web/foam */
        edge *copyTag;
        /* temporary, used in glueFoam */
        bool toBeDeleted;

        /* copyTag must always be reset to 0 before use */
        edge() : copyTag(0), toBeDeleted(false) {}

        oEdge* asOEdge() { return (oEdge*)(this); }
        iEdge* asIEdge() { return (iEdge*)(this); }
};

class oEdge : public edge {
    public:
        /* from, to. Are 0 if and only if the edge is a circle. */
        point *points[2]; 

        /* temporary, points to the facet this edge is adjacent to when a copy
         * of the whole web together with a foam from the web to its copy is
         * created.
         */
        facet *idFacet;
        /* temporary, marks if the edge belongs to the domain or the range
         * web of a foam.
         */
        bool isBelow;

        oEdge();

        iPoint* pointAsI(int i) const { return (iPoint *)(points[i]); }
        bool isCircle() const { return (!points[0]); }
};

class iPoint : public point {
    public:
        /* the three adjacent edges in the web, in counterclockwise order, if
         * the points is not a transit-point; otherwise, edges[2] == 0, and
         * edges[0] is the incoming, edges[1] the outgoing edge (or the other
         * way(?)
         */
        oEdge *edges[3];

        iEdge *idSingular;
        iPoint *copyTag;
        oEdge *oldEdges[3];

        iPoint();

        void saveEdges();
        void restoreEdges();
        bool isTransit() const;
        bool isIncoming() const;
        oEdge* otherEdge(const oEdge &oneEdge) const;
        /* returns the other iPoints which are not transit point (max. 3) this
         * is connected to via the oEdges in w.
         * goal must be an array of size 3; if the i-th edge of iPoint goes to
         * another non-transit iPoint, it is written to the i-th entry of the
         * array, otherwise 0.
         */
        void findGoals(iPoint* goal[], std::vector<oEdge*> w[]) const;
};

class facet {
    public:
        unsigned short dots;
        BOUNDARIESCONT<EDGECONT<edge*> > boundaries;

        facet* copyTag;
        facet* heir;

        facet() : dots(0), copyTag(0), heir(0) { }

        bool isEquivalentTo(const facet& other) const;
        void sortOneBoundary(EDGECONT<edge*> &L);
        void sort();
        bool mayBePartOfIdentity() const;
        void eraseEdge(const iEdge &e);
        bool isSphere() const { return boundaries.empty(); }
};

class iEdge : public edge {
    public:
        iPoint *points[2]; 
        facet *facets[3];

        iEdge *heir;

        bool isCircle() const { return (!points[0]); }

        iEdge() : heir(0) { isInner = true; }
};



class web {
    public:
        short qShift;
        IPOINTCONT<iPoint*> points;
        OEDGECONT<oEdge*> edges;

        web() : qShift(0) { }

        void setEdgesToBelow() const;
        void setEdgesToAbove() const;
        void composeWeb(web *other);
        bool find0Gon(web* replacement[], std::vector<foam*> &foamRi,
                std::vector<foam*> &foamLe) const;
        bool find2Gon(web* replacement[], std::vector<foam*> &foamRi,
                std::vector<foam*> &foamLe) const;
        bool find4Gon(web* replacement[], std::vector<foam*> &foamRi,
                std::vector<foam*> &foamLe) const;
        int findNGon(web* replacement[], std::vector<foam*> &foamRi,
                std::vector<foam*> &foamLe) const;
        void deleteContents();
        web* copy(foam* id = 0);
};

class foam {
    public:
        ratio coefficient;
        FACETCONT<facet*> facets;
        SINGULARCONT<iEdge*> singulars;

        foam() { }
        explicit foam(ratio val) : coefficient(val) { }

        void sort();
        bool isMultipleOf(const foam& other) const;
        bool isInt() const;
        void inverseCoefficient() { coefficient.inv(); }
        void changeSign() { coefficient *= ratio(-1); }
        bool dualEqualsInverse() const;
        void glueFoam(foam *other);
        void composeFoam(foam *other, const web &below, const web &below2);
        foam* copy(bool dualizeIt = false);
        foam* copyNoDual() { return copy(); }
        bool isEmpty() const { return facets.empty(); }
        foam* dualize();
        void deleteContents();
        int degree();
};

class lcfoams {
    public:
        FOAMCONT<foam*> foams;

        bool doGlueSS;

        lcfoams() : doGlueSS(false) { }
        explicit lcfoams(foam* aSingleFoam);

        ratio sumOfCoefficients() const;
        bool dualEqualsInverse(web &below) const;
        void kill() { deleteContents(); delete this; }
        void deleteContents();
        bool isInt() const;
        bool isZero() const;
        void glueLcfoams(lcfoams* other, bool dualizeIt = false,
                bool deleteIt = true);
        void composeLcFoams(foam *other, const web & below, const web & below2,
                bool deleteIt = true);
        void glueLcfoams(foam *other, bool dualizeIt = false,
                bool deleteIt = true);
        void changeSign();
        void dualize();
        lcfoams* copy(bool dualizeIt = false) const; 
        lcfoams* copyNoDualAndDeleteZero();
        foam* firstFoam() const { return *(foams.begin()); }
        bool simplify();
};

class one {
    public:
        WEBCONT<web*> webs;

        void eraseWeb(unsigned idx);
        int oneSimplify(two *twoLeft, two *twoRight, bool ini,
                std::deque<std::vector<unsigned> > *gaussStack,
                unsigned leftIdx, bool markedReduced);
        unsigned realSize() const;
        void deleteContents();
};

class two : public SparseMatrix<lcfoams> {
    public:
        void deleteContents();
        bool lcfSimplify(MatrixIterator<lcfoams> *it);
        void checkForZeroes();

        two(foam *f = 0);
};

class chainComplex {
    public:
        std::vector<one*> spaces;
        std::vector<two*> maps;
        int tShift;
        int qShift;
        void computeHomology(bool rational, unsigned &QIdx, int &minQ,
                std::vector<std::vector<unsigned> > &homDims,
                std::vector<torstruct> &torsion);

        chainComplex() : tShift(0), qShift(0) {};

        bool gauss(unsigned mapIdx, unsigned rowIdx,
                unsigned colIdx,
                std::deque<std::vector<unsigned> > *gaussStack, bool check);
        void composeEverything(web* W) { composeEverything(W, W, 0); }
        void composeEverything(web* W1, web* W2, foam* map);
        void tensor(web* W1, web* W2, foam* map);
        void setToSmall();
        void tensorLine(oPoint* in, oPoint* out);
        void tensorPosCros(oPoint* loIn, oPoint* hiOut, oPoint* loOut,
                oPoint* hiIn);
        void tensorNegCros(oPoint* loIn, oPoint* hiIn, oPoint* loOut,
                oPoint* hiOut);
        void getMinMaxQ(int &minQ, int &maxQ);
        void deleteContents();
};
