/*
 *    output.C --- Wraps some of the output.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <vector>
#include <deque>
#include <list>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"
#include "output.H"

unsigned printHomologyModN(unsigned int ccSpacesSize, int tShift, unsigned QIdx,
        int shiftedMinQ, std::vector<std::vector<unsigned int> > &homDims);

void printHomology(unsigned int ccSpacesSize, int tShift, unsigned int QIdx,
        int shiftedMinQ, std::vector<std::vector<unsigned int> > &homDims,
        std::vector<torstruct> torsion, bool rational, bool reduced) {
    std::cout << "Rational " << (reduced ? "reduced " : "") << "homology: ";
    unsigned totalDim = 
        printHomologyModN(ccSpacesSize, tShift, QIdx, shiftedMinQ, homDims);
    std::cout << "Total rank: " << totalDim << "\n";
    if (! rational) {
        if (torsion.empty())
            std::cout << "The homology is torsion-free.\n";
        for (std::vector<torstruct>::iterator vtit = torsion.begin();
                vtit != torsion.end(); ++vtit) {
            std::cout << vtit->mod << "-torsion: ";
            printHomologyModN(ccSpacesSize, tShift, QIdx,
                    shiftedMinQ, vtit->dim);
        }
    }
}

unsigned printHomologyModN(unsigned int ccSpacesSize, int tShift, unsigned QIdx,
        int shiftedMinQ, std::vector<std::vector<unsigned int> > &homDims) {
    unsigned result = 0;
    bool isFirst = true;
    for (unsigned int i = 0; i < ccSpacesSize; ++i) {
        bool isFirst2 = true;
        for (unsigned int j = 0; j < QIdx; ++j) {
            const unsigned int dim = homDims.at(i).at(j);
            if (dim != 0) {
                result += dim;
                if (isFirst2) {
                    isFirst2 = false;
                    if (isFirst) isFirst = false;
                    else std::cout << " + ";
                    switch ((int)(i) + tShift) {
                        case 0:  break;
                        case 1:  std::cout << "t"; break;
                        default:  std::cout << "t^" << (int)(i) + tShift;
                    }
                    std::cout << "(";
                } else std::cout << " + ";
                if (dim != 1) std::cout << dim;
                switch ((int)(2 * j) + shiftedMinQ) {
                    case 0:  if (dim == 1) std::cout << "1"; break;
                    case 1:  std::cout << "q"; break;
                    default:  std::cout << "q^" << (int)(2 * j) + shiftedMinQ;
                }
            }
        }
        if (! isFirst2) std::cout << ")";
    }
    std::cout << "\n";
    return result;
}

void printRasmussen(unsigned s, int shift) {
    if (s != NO_RASMUSSEN)
        std::cout << "s_3-concordance invariant: " << (int)(s) + shift << "\n";
    else
        std::cout << "Cannot get s3-concordance invariant - either you "
            "entered a multi-component link, or Lee's spectral sequence does "
            "not converge on the second page.\n";
}

void printVVi(VECVECINT &v) {
    for (VECVECINT::iterator vviIt = v.begin(); vviIt != v.end(); ++vviIt) {
        if (vviIt != v.begin()) std::cout << ",";
        std::cout << "[";
        for (VECINT::iterator viIt = vviIt->begin();
                viIt + 1 != vviIt->end(); ++viIt) {
            if (viIt != vviIt->begin()) std::cout << ",";
            std::cout << *viIt + 1;
        }
        std::cout << "]";
    }
    std::cout << "].\n";
}

void printTitle()  {
    std::cout << "foamho, a sl3-homology calculator, version " << VERSION <<
        ". Have fun. Call with \"-h\" for help.\n";
}

void printHelp() {
    std::cout << "foamho, a sl3-homology calculator, version " << VERSION
        << ".\n\n" << "Usage: foamho [OPTIONS] braid | pd | dt NOTATION\n\n"
        << "For example, the following three commands all compute the\n"
        << "figure-8-knot's integral homology:\n\n"
        << "                    foamho braid aBaB\n"
        << "                    foamho pd [[4,2,5,1],[8,6,1,5],[6,3,7,4],"
        "[2,7,3,8]]\n"
        << "                    foamho dt [4,6,8,2]\n\n"
        << "Options:\n"
        << "-q                  Compute rational homology instead of "
                                                                   "integral.\n"
        << "-r                  Compute reduced homology instead of unreduced. "
                                                                "You may give\n"
        << "                    a number right after -r to indicate the marked"
                                                                     " strand\n"
        << "                    (useful for links).\n\n"
        << "-g                  Do not attempt to optimise the girth.\n"
        << "-v                  Display some progress information.\n"
        << "-vv                 Display more detailed progress information.\n"
        << "-t                  Display time and memory consumption.\n"
        /*<< "-rr                 Compute both reduced and unreduced
         * homology.\n\n" \ */
        << "-h                  Display this help message and exit.\n\n"
        << "Written in 2012 by Lukas Lewark, lewark@math.jussieu.fr. "
                                                   "All feedback is welcome.\n";
}
