/*
 *    simplifyWebs.C --- Contains the functions that simplify webs by
 *                       looking for circles, digons and squares, and
 *                       removing them using the appropriate isomorphism-foams.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <iostream>
#include <list>
#include <vector>
#include <deque>
#include <string.h>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"

int one::oneSimplify(two *twoLeft, two *twoRight, bool ini,
        std::deque<std::vector<unsigned> > *gaussStack, unsigned leftIdx,
        bool markedReduced) {
    /* 1. look for circle / digon / square and produce the corresponding
     *    foams in both directions
     * 2. make copies of rows / columns in matrices
     * 3. compose the whole row / column with that foam
     * 4. remove the circle / digon / square from the web, and replace the web
     *     by two or three (shifted) copies
     */
    static unsigned int idx;

    if (ini) {
        idx = 0;
        return -1;
    }

    for (; idx < webs.size(); ++idx) {
        web* thisWeb = webs.at(idx);
        if (! thisWeb) continue;
        web* replacement[3];
        std::vector<foam*> foamRi, foamLe;
        int whichNGon;
        if ( (whichNGon = thisWeb->findNGon(replacement, foamRi, foamLe)) ) {
            if (markedReduced) {
                if (whichNGon != 1) throw 34;
                replacement[0] = replacement[1];
                replacement[1] = 0;
                replacement[2] = 0;
                foamRi.pop_back();
                foamRi.pop_back();
                foamLe.pop_back();
                foamLe.pop_back();
            }
            const int numCopies = foamRi.size();
            for (int i = 0; i < numCopies; ++i)
                webs.push_back(replacement[i]);
            if (twoRight) {
                twoRight->copyCols(idx, numCopies,
                        &lcfoams::copyNoDualAndDeleteZero);
                for (int i = 0; i < numCopies; ++i) {
                    const unsigned int c = twoRight->colsSize() - numCopies + i;
                    ENTRYCONT<SparseEntry<lcfoams> > * L = twoRight->colsSM(c);
                    ENTRYCONT<SparseEntry<lcfoams> >::iterator j = L->begin();
                    while ( j != L->end()) {
                        j->val->glueLcfoams(foamRi.at(i), false, false);
                        if (j->val->simplify()) {
                            j->val->deleteContents();
                            delete j->val;
                            twoRight->erase(j->idx, c, 0, &j);
                        } else {
                            if (gaussStack) 
                                if (j->val->dualEqualsInverse(*webs.at(c))) {
                                    std::vector<unsigned> v;
                                    v.push_back(leftIdx + 1);
                                    v.push_back(j->idx);
                                    v.push_back(c);
                                    gaussStack->push_back(v);
                                }
                            ++j;
                        }
                    }
                }
                twoRight->eraseCol(idx, &lcfoams::kill);
            }
            if (twoLeft) {
                twoLeft->copyRows(
                        idx, numCopies, &lcfoams::copyNoDualAndDeleteZero);
                for (int i = 0; i < numCopies; ++i) {
                    const unsigned int r = twoLeft->rowsSize() - numCopies + i;
                    ENTRYCONT<SparseEntry<lcfoams> > * L = twoLeft->rowsSM(r);
                    ENTRYCONT<SparseEntry<lcfoams> >::iterator j = L->begin();
                    while ( j != L->end()) {
                        j->val->glueLcfoams(foamLe.at(i), false, false);
                        j->val->simplify();
                        if (j->val->simplify()) {
                            j->val->deleteContents();
                            delete j->val;
                            twoLeft->erase(r, j->idx, &j, 0);
                        } else {
                            if (gaussStack) 
                                if (j->val->dualEqualsInverse(*webs.at(r))) {
                                    std::vector<unsigned> v;
                                    v.push_back(leftIdx);
                                    v.push_back(r);
                                    v.push_back(j->idx);
                                    gaussStack->push_back(v);
                                }
                            ++j;
                        }
                    }
                }
                twoLeft->eraseRow(idx, &lcfoams::kill);
            }
            for (int i = 0; i < numCopies; ++i) {
                foamRi.at(i)->deleteContents();
                delete foamRi.at(i);
                foamLe.at(i)->deleteContents();
                delete foamLe.at(i);
            }
            webs.at(idx) = 0;
            thisWeb->deleteContents();
            delete thisWeb;
            return whichNGon;
        }
    }
    return -1;
}

bool web::find0Gon(web* replacement[], std::vector<foam*> &foamRi,
        std::vector<foam*> &foamLe) const {
    OEDGECONT<oEdge*> v;
    for (OEDGECONT<oEdge*>::const_iterator i = edges.begin();
            i != edges.end(); ++i) {
        if ((*i)->isCircle()) {
            v.push_back(*i);
            break;
        } else {
            oEdge *j = *i;
            for (;;) {
                v.push_back(j);
                if (j->points[1] == (*i)->points[0]) break;
                if (j->points[1]->isInner) {
                    iPoint *p = (iPoint *)(j->points[1]);
                    if (p->isTransit()) j = p->otherEdge(*j);
                    else break;
                } else break;
            }
            if (j->points[1] == (*i)->points[0]) break;
            else v.clear();
        }
    }
    if (! v.empty()) {

        web wcopy = *this; /* shallow copy */
        if (v.size() == 1)
            wcopy.edges.remove(*(v.begin()));
        else for (OEDGECONT<oEdge*>::const_iterator i = v.begin();
                i != v.end(); ++i) {
            wcopy.points.remove((*i)->pointAsI(0));
            wcopy.edges.remove(*i);
        }
        foamLe.resize(3);
        foamRi.resize(3);
        for (int i = 0; i < 3; ++i) {
            foamLe.at(i) = new foam;
            replacement[i] = wcopy.copy(foamLe.at(i));
            replacement[i]->qShift += (- 2 * i + 2);

            foam *f = foamLe.at(i);
            f->coefficient = ratio(-1);
            f->facets.resize(f->facets.size() + 1);
            facet *fa = new facet();
            *(f->facets.rbegin()) = fa;
            fa->dots = i;
            fa->boundaries.resize(1);
            EDGECONT<edge*>* l = &*(fa->boundaries.begin());
            for (OEDGECONT<oEdge*>::const_iterator itoe = v.begin();
                    itoe != v.end(); ++itoe)
                l->push_back(*itoe);

            foam *f2 = ( foamRi.at(i) = f->copy(true) );
            f2->coefficient = ratio(1);
            facet *fa2 = fa->copyTag;
            fa2->dots = 2 - fa2->dots;
        }
        return true;
    }
    return false;
}

bool web::find2Gon(web* replacement[], std::vector<foam*> &foamRi,
        std::vector<foam*> &foamLe) const {
    OEDGECONT<oEdge*> v1, v2;
    iPoint* vertices[2];
    vertices[0] = vertices[1] = 0;
    for (IPOINTCONT<iPoint*>::const_iterator i = points.begin();
            i != points.end(); ++i) {
        if ((*i)->isTransit() || ((*i)->isIncoming())) continue;
        iPoint* goal[3];
        memset(goal, 0, sizeof goal);
        std::vector<oEdge*> w[3];
        (*i)->findGoals(goal, w);
        for (int j = 0; j < 3; ++j)
            if ((goal[j] == goal[(j + 1) % 3]) && (goal[j])) {
                vertices[0] = *i;
                vertices[1] = goal[j];
                v1.insert(v1.begin(), w[j].begin(), w[j].end());
                v2.insert(v2.begin(), w[(j + 1) % 3].rbegin(),
                        w[(j + 1) % 3].rend());

                break;
            }
        if (! v1.empty()) break;
    }
    if (! v1.empty()) {

        web wcopy = *this; /* shallo copy */

        {OEDGECONT<oEdge*>::iterator i = v1.begin();
        while (i != v2.end()) {
            if (i == v1.end()) i = v2.begin();
            for (int k = 0; k < 2; ++k)
                (*i)->pointAsI(k)->saveEdges();
            ++i;
        }}

        {OEDGECONT<oEdge*>::iterator i = v1.begin();
        while (i != v2.end()) {
            if (i == v1.end())
                i = v2.begin();
            iPoint *toRemove = (*i)->pointAsI(0);
            if (toRemove != vertices[0])
                wcopy.points.remove(toRemove);
            wcopy.edges.remove(*i);
            for (int k = 0; k < 2; ++k) {
                for (int j = 0; j < 3; ++j)
                    if ((*i)->pointAsI(k)->edges[j] == *i) 
                        (*i)->pointAsI(k)->edges[j] = 0;
            }
            ++i;
        }}

        oEdge *newe = new oEdge();
        newe->points[0] = vertices[1];
        newe->points[1] = vertices[0];

        for (int k = 0; k < 2; ++k)
            for (int j = 0; j < 3; ++j)
                if ((j != 1 - k) && (vertices[k]->edges[j])) {
                    vertices[k]->edges[1 - k] = vertices[k]->edges[j];
                    vertices[k]->edges[j] = 0;
                    break;
                }
        vertices[1]->edges[1] = newe;
        vertices[0]->edges[0] = newe;
        wcopy.edges.push_back(newe);
        foamLe.resize(2);
        foamRi.resize(2);
        for (int i = 0; i < 2; ++i) {
            foamLe.at(i) = new foam();
            replacement[i] = wcopy.copy(foamLe.at(i));
            replacement[i]->qShift += (1 - 2*i);

            foam *f = foamLe.at(i);
            f->coefficient = ratio(1 - 2 * i);
            iEdge *newie = new iEdge();
            newie->points[0] = vertices[1];
            newie->points[1] = vertices[0];
            facet *fa1 = new facet();
            facet *fa2 = new facet();
            fa2->dots = i;
            fa1->dots = 0;
            fa1->boundaries.resize(1);
            fa2->boundaries.resize(1);
            {
                EDGECONT<edge*>* l = &*(fa1->boundaries.begin());
                l->insert(l->begin(), v1.begin(), v1.end());
                l->push_back(newie);
            } {
                EDGECONT<edge*>* l = &*(fa2->boundaries.begin());
                l->insert(l->begin(), v2.rbegin(), v2.rend());
                l->push_back(newie);
            }
            facet *fa3 = newe->idFacet;
            bool foundIt = false;
            for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                    fa3->boundaries.begin(); ile != fa3->boundaries.end();
                    ++ile) {
                for (EDGECONT<edge*>::iterator ie = ile->begin();
                        ie != ile->end(); ++ie)
                    if (*ie == newe) {
                        ile->insert(ie, newie);
                        ile->erase(ie);
                        foundIt = true;
                        break;
                    }
                if (foundIt) break;
            }
            if (!foundIt) throw 16;
            newie->facets[0] = fa1;
            newie->facets[1] = fa2;
            newie->facets[2] = fa3;
            f->facets.push_back(fa1);
            f->facets.push_back(fa2);
            f->singulars.push_back(newie);

            foamRi.at(i) = f->copy(true);
            fa2->copyTag->dots = 0;
            fa1->copyTag->dots = 1 - i;
            foamRi.at(i)->coefficient = ratio(1);
        }
        replacement[2] = 0;

        delete newe;

        {OEDGECONT<oEdge*>::const_iterator i = v1.begin();
        while (i != v2.end()) {
            if (i == v1.end()) i = v2.begin();
            for (int k = 0; k < 2; ++k)
                (*i)->pointAsI(k)->restoreEdges();
            ++i;
        }}

        if (!((foamLe.at(0)->degree() == -1) && (foamLe.at(1)->degree() == 1)
                    && (foamRi.at(0)->degree() == 1)
                    && (foamRi.at(1)->degree() == -1))) {
            std::cerr << "digon degree error.\n";
            throw 17;
        }
        return true;
    }
    return false;
}

bool web::find4Gon(web* replacement[], std::vector<foam*> &foamRi,
        std::vector<foam*> &foamLe) const {
    OEDGECONT<oEdge*> v;
    iPoint* vertices[4];
    for (IPOINTCONT<iPoint*>::const_iterator i = points.begin();
            i != points.end(); ++i) {
        if ((*i)->isTransit() || ((*i)->isIncoming())) continue;
        iPoint* goal[3];
        memset(goal, 0, sizeof goal);
        std::vector<oEdge*> w[3];
        iPoint* goal2[3][3];
        memset(goal2, 0, sizeof goal2);
        std::vector<oEdge*> w2[3][3];
        (*i)->findGoals(goal, w);
        for (int j = 0; j < 3; ++j)
            if (goal[j])
                goal[j]->findGoals(goal2[j], w2[j]);
        OEDGECONT<oEdge*>::iterator vSplitIt1, vSplitIt2, vSplitIt3;

        int jk[3];
        for (int y = 0; y < 3; ++y) {
            jk[y] = -1;
            if (! goal[y]) continue;
            for (int x = 0; x < 3; ++x)
                if (goal2[y][x] == *i)
                    jk[y] = x;
            if (jk[y] == -1) throw 18;
        }
        for (int j = 0; j < 3; ++j) {
            const int j2 = (j + 1) % 3;
            if ((! goal[j]) || (! goal[j2])) continue;
            const int k = (jk[j] + 1) % 3;
            if (k == jk[j]) continue;
            const int k2 = (jk[j2] + 2) % 3;

            if ((goal[j] == goal[(j + 2) % 3]) ||
                    (goal[j2] == goal[(j + 2) % 3]) ||
                    (goal2[j][(k + 1) % 3] == goal2[j][k]) ||
                    (goal2[j2][(k2 + 2) % 3] == goal2[j][k])) continue;

            if ((goal2[j][k] == goal2[j2][k2]) && (goal2[j][k]) &&
                    (goal2[j][k] != *i) && (goal[j] != goal[j2])) {
                vertices[0] = *i;
                vertices[1] = goal[j];
                vertices[2] = goal2[j][k];
                vertices[3] = goal[j2];
                v.insert(v.begin(), w[j].begin(), w[j].end());
                vSplitIt1 = --v.end();
                v.insert(v.end(), w2[j][k].rbegin(), w2[j][k].rend());
                vSplitIt2 = --v.end();
                v.insert(v.end(), w2[j2][k2].rbegin(), w2[j2][k2].rend());
                vSplitIt3 = --v.end();
                v.insert(v.end(), w[j2].begin(), w[j2].end());
                ++vSplitIt1;
                ++vSplitIt2;
                ++vSplitIt3;
                break;
            }
        }
        if (! v.empty()) {
            web wcopy = *this; /* shallow copy */
            for (OEDGECONT<oEdge*>::const_iterator it = v.begin();
                    it != v.end(); ++it)
                for (int k = 0; k < 2; ++k)
                    (*it)->pointAsI(k)->saveEdges();

            for (OEDGECONT<oEdge*>::const_iterator it = v.begin();
                    it != v.end(); ++it) {
                iPoint *toRemove = (*it)->pointAsI(0);
                if ((toRemove != vertices[0]) && (toRemove != vertices[2]))
                    wcopy.points.remove(toRemove);
                wcopy.edges.remove(*it);
                for (int k = 0; k < 2; ++k)
                    for (int j = 0; j < 3; ++j) {
                        if ((*it)->pointAsI(k)->edges[j] == *it)
                            (*it)->pointAsI(k)->edges[j] = 0;
                    }
            }

            for (int k = 0; k < 4; ++k)
                for (int j = 0; j < 3; ++j)
                    if ((j != (1 - (k % 2))) && (vertices[k]->edges[j])) {
                        vertices[k]->edges[1 - (k % 2)] = vertices[k]->edges[j];
                        vertices[k]->edges[j] = 0;
                        break;
                    }
            web wcopy2 = wcopy; /* shallow copy */

            {
                oEdge *newe = new oEdge();
                oEdge *newe2 = new oEdge();
                newe->points[0] = vertices[1];
                newe->points[1] = vertices[0];
                vertices[1]->edges[1] = newe;
                vertices[0]->edges[0] = newe;
                newe2->points[0] = vertices[3];
                newe2->points[1] = vertices[2];
                vertices[3]->edges[1] = newe2;
                vertices[2]->edges[0] = newe2;
                wcopy.edges.push_back(newe);
                wcopy.edges.push_back(newe2);

                foamLe.resize(2);
                foamRi.resize(2);
                foamLe.at(0) = new foam();
                replacement[0] = wcopy.copy(foamLe.at(0));
                {
                    foam *f = foamLe.at(0);
                    f->coefficient = ratio(-1);
                    iEdge *newie1 = new iEdge();
                    iEdge *newie2 = new iEdge();
                    newie1->points[0] = vertices[1];
                    newie1->points[1] = vertices[0];
                    newie2->points[0] = vertices[3];
                    newie2->points[1] = vertices[2];
                    facet *fa1 = new facet();
                    facet *fa2 = new facet();
                    facet *fa3 = new facet();
                    fa1->dots = 0;
                    fa2->dots = 0;
                    fa3->dots = 0;
                    fa1->boundaries.resize(1);
                    fa2->boundaries.resize(1);
                    fa3->boundaries.resize(1);
                    {
                        EDGECONT<edge*>* l = &*(fa1->boundaries.begin());
                        l->insert(l->begin(), v.begin(), vSplitIt1);
                        l->push_back(newie1);
                    } {
                        EDGECONT<edge*>* l = &*(fa2->boundaries.begin());
                        l->insert(l->begin(), vSplitIt1, vSplitIt2);
                        l->push_back(newie1);
                        l->insert(l->end(), vSplitIt3, v.end());
                        l->push_back(newie2);
                    } {
                        EDGECONT<edge*>* l = &*(fa3->boundaries.begin());
                        l->insert(l->begin(), vSplitIt2, vSplitIt3);
                        l->push_back(newie2);
                    } 
                    facet *fa4 = newe->idFacet;
                    bool foundIt = false;
                    for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                            fa4->boundaries.begin();
                            ile != fa4->boundaries.end(); ++ile) {
                        for (EDGECONT<edge*>::iterator ie = ile->begin();
                                ie != ile->end(); ++ie)
                            if (*ie == newe) {
                                ile->insert(ie, newie1);
                                ile->erase(ie);
                                foundIt = true;
                                break;
                            }
                        if (foundIt) break;
                    }
                    if (!foundIt) throw 19;
                    foundIt = false;
                    facet *fa5 = newe2->idFacet;
                    for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                            fa5->boundaries.begin();
                            ile != fa5->boundaries.end(); ++ile) {
                        for (EDGECONT<edge*>::iterator ie = ile->begin();
                                ie != ile->end(); ++ie)
                            if (*ie == newe2) {
                                ile->insert(ie, newie2);
                                ile->erase(ie);
                                foundIt = true;
                                break;
                            }
                        if (foundIt) break;
                    }
                    if (!foundIt) throw 20;
                    newie1->facets[0] = fa4;
                    newie1->facets[1] = fa1;
                    newie1->facets[2] = fa2;

                    newie2->facets[0] = fa3;
                    newie2->facets[1] = fa2;
                    newie2->facets[2] = fa5;

                    f->facets.push_back(fa1);
                    f->facets.push_back(fa2);
                    f->facets.push_back(fa3);
                    f->singulars.push_back(newie1);
                    f->singulars.push_back(newie2);
                }
                delete newe;
                delete newe2;
            }
            {
                oEdge *newe = new oEdge();
                oEdge *newe2 = new oEdge();
                newe->points[0] = vertices[1];
                newe->points[1] = vertices[2];
                vertices[1]->edges[1] = newe;
                vertices[2]->edges[0] = newe;
                newe2->points[0] = vertices[3];
                newe2->points[1] = vertices[0];
                vertices[3]->edges[1] = newe2;
                vertices[0]->edges[0] = newe2;
                wcopy2.edges.push_back(newe);
                wcopy2.edges.push_back(newe2);
                foamLe.at(1) = new foam();


                replacement[1] = wcopy2.copy(foamLe.at(1));
                replacement[2] = 0;
                {
                    foam *f = foamLe.at(1);
                    f->coefficient = ratio(1);
                    iEdge *newie1 = new iEdge();
                    iEdge *newie2 = new iEdge();
                    newie1->points[0] = vertices[1];
                    newie1->points[1] = vertices[2];
                    newie2->points[0] = vertices[3];
                    newie2->points[1] = vertices[0];
                    facet *fa1 = new facet();
                    facet *fa2 = new facet();
                    facet *fa3 = new facet();
                    fa1->dots = 0;
                    fa2->dots = 0;
                    fa3->dots = 0;
                    fa1->boundaries.resize(1);
                    fa2->boundaries.resize(1);
                    fa3->boundaries.resize(1);
                    {
                        EDGECONT<edge*>* l = &*(fa1->boundaries.begin());
                        l->insert(l->begin(), vSplitIt1, vSplitIt2);
                        l->push_back(newie1);
                    } {
                        EDGECONT<edge*>* l = &*(fa2->boundaries.begin());
                        l->insert(l->begin(), v.begin(), vSplitIt1);
                        l->push_back(newie1);
                        l->insert(l->end(), vSplitIt2, vSplitIt3);
                        l->push_back(newie2);
                    } {
                        EDGECONT<edge*>* l = &*(fa3->boundaries.begin());
                        l->insert(l->end(), vSplitIt3, v.end());
                        l->push_back(newie2);
                    } 
                    facet *fa4 = newe->idFacet;
                    bool foundIt = false;
                    for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                            fa4->boundaries.begin();
                            ile != fa4->boundaries.end(); ++ile) {
                        for (EDGECONT<edge*>::iterator ie = ile->begin();
                                ie != ile->end(); ++ie)
                            if (*ie == newe) {
                                ile->insert(ie, newie1);
                                ile->erase(ie);
                                foundIt = true;
                                break;
                            }
                        if (foundIt) break;
                    }
                    if (!foundIt) throw 21;
                    foundIt = false;
                    facet *fa5 = newe2->idFacet;
                    for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                            fa5->boundaries.begin();
                            ile != fa5->boundaries.end(); ++ile) {
                        for (EDGECONT<edge*>::iterator ie = ile->begin();
                                ie != ile->end(); ++ie)
                            if (*ie == newe2) {
                                ile->insert(ie, newie2);
                                ile->erase(ie);
                                foundIt = true;
                                break;
                            }
                        if (foundIt) break;
                    }
                    if (!foundIt) throw 22;
                    newie1->facets[0] = fa2;
                    newie1->facets[1] = fa1;
                    newie1->facets[2] = fa4;

                    newie2->facets[0] = fa5;
                    newie2->facets[1] = fa2;
                    newie2->facets[2] = fa3;

                    f->facets.push_back(fa1);
                    f->facets.push_back(fa2);
                    f->facets.push_back(fa3);
                    f->singulars.push_back(newie1);
                    f->singulars.push_back(newie2);
                }

                for (OEDGECONT<oEdge*>::const_iterator it = v.begin();
                        it != v.end(); ++it)
                    for (int k = 0; k < 2; ++k)
                        (*it)->pointAsI(k)->restoreEdges();

                delete newe;
                delete newe2;
            }
            foamRi.resize(2);
            for (int k = 0; k < 2; ++k) {
                foamRi.at(k) = foamLe.at(k)->copy(true);
                foamRi.at(k)->coefficient *= ratio(-1);
            }
            if (!((foamLe.at(0)->degree() == 0) &&
                        (foamLe.at(1)->degree() == 0) &&
                        (foamRi.at(0)->degree() == 0) &&
                        (foamRi.at(1)->degree() == 0))) {
                std::cerr << "square degree error.\n";
                throw 23;
            }
            return true;
        }
    }
    return false;
}

int web::findNGon(web* replacement[], std::vector<foam*> &foamRi,
        std::vector<foam*> &foamLe) const {
    int result = 0;
    if (find0Gon(replacement, foamRi, foamLe)) result = 1;
    else if (find2Gon(replacement, foamRi, foamLe)) result = 2;
    else if (find4Gon(replacement, foamRi, foamLe)) result = 3;
    return result;
}
