/*
 *    smith.C --- Wrapper for computing the smith normal form using pari
 *                (pari function "ZM_snf").
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <iostream>
#include <vector>
#include <list>
#include <deque>
#include <pari/pari.h>
#include <mpir.h>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"
#include "smith.H"

const long unsigned PariStackSize = 50000000LU;

bool withPari();
void initPari();
GEN mpz2GEN(mpz_t X);
void smithNormalForm(chainComplex *CC, std::vector<torstruct> &torsion,
        std::vector<std::vector<unsigned> > &ranks,
        std::vector<std::vector<unsigned int> > ccDims,
        std::vector<std::vector<unsigned int> > IdxTranslator,
        unsigned int ccSpacesSize, unsigned QIdx, unsigned minQ);

bool withPari() {
    return true;
}

void initPari() {
    pari_init(PariStackSize, 0);
}

GEN mpz2GEN(mpz_t X) {
    long l = X->_mp_size;
    unsigned long lx = labs(l)+2;
    GEN x = cgeti(lx);
    x[1] = evalsigne(l > 0? 1: -1) | evallgefint(lx);
    for (unsigned i = 2; i < lx; i++) x[i] = X->_mp_d[i-2];
    return x;
}

void smithNormalForm(chainComplex *CC, std::vector<torstruct> &torsion,
        std::vector<std::vector<unsigned> > &ranks,
        std::vector<std::vector<unsigned int> > ccDims,
        std::vector<std::vector<unsigned int> > IdxTranslator,
        unsigned int ccSpacesSize, unsigned QIdx, unsigned minQ) {
    initPari();
    for (unsigned int i = 0; i < CC->maps.size(); ++i) {
        std::vector<GEN> vSM;
        vSM.reserve(QIdx);
        for (unsigned int j = 0; j < QIdx; ++j)
            vSM.push_back(zeromatcopy(ccDims.at(i + 1).at(j),
                        ccDims.at(i).at(j)));

        for (MatrixIterator<lcfoams> mit = CC->maps.at(i)->begin();
                ! mit.isFinished(); mit.step()) {
            web* const rowWeb = CC->spaces.at(i + 1)->webs.at(mit.r());
            web* const colWeb = CC->spaces.at(i)->webs.at(mit.c());
            const int qOnRight = (rowWeb->qShift - minQ) / 2;
            const int qOnLeft  = (colWeb->qShift - minQ) / 2;
            if (qOnLeft == qOnRight) {
                GEN mft = mpz2GEN(*((mpz_t*)(mit->sumOfCoefficients().val)));
                gcoeff(vSM.at(qOnLeft), 1 + IdxTranslator.at(i + 1).at(mit.r()),
                        1 + IdxTranslator.at(i).at(mit.c())) = mft;
            }
            else if (*mit != 0) {
                std::cerr << "Error: Non-zero map between two spaces of "
                    "non-equal degree, row " << mit.r() << ", col " <<
                    mit.c() << ".\n";
                throw 33;
            }
        }

        for (unsigned int j = 0; j < QIdx; ++j) {
            GEN theMatrix = vSM.at(j);
            if ((ccDims.at(i).at(j) > 0) && (ccDims.at(i + 1).at(j) > 0)) {
                unsigned r = 0;
                GEN y = ZM_snf(theMatrix);

                std::vector<unsigned> divisorIdx;
                unsigned prevF = 1;
                for (unsigned k = lg(y) - 1; (k >= 1) &&
                        (signe(gel(y, k)) != 0); --k) {
                    ++r;
                    unsigned f = itos(gel(y, k));
                    unsigned quot = f / prevF;
                    GEN factors = Z_factor(stoi(quot));
                    for (unsigned x = 1; x < lg(gel(factors,1)); ++x) {
                        unsigned primePowerFactor =
                            itos(powii(gmael2(factors, 1, x),
                                        gmael2(factors, 2, x)));

                        unsigned newPrimePower = primePowerFactor;
                        std::vector<unsigned>::iterator toBeErased =
                            divisorIdx.end();
                        for (std::vector<unsigned>::iterator vit =
                                divisorIdx.begin();
                                vit != divisorIdx.end(); ++vit) {
                            const unsigned oldPrimePower =
                                torsion.at(*vit).mod;
                            if (primePowerFactor % oldPrimePower == 0) {
                                toBeErased = vit;
                                newPrimePower *= oldPrimePower;
                            } else
                                torsion.at(*vit).dim.at(i + 1).at(j) += 1;
                        }
                        if (toBeErased != divisorIdx.end())
                            divisorIdx.erase(toBeErased);
                        if (newPrimePower > 1) {
                            unsigned newIdx = divisorIdx.size();
                            for (unsigned l = 0; l < divisorIdx.size(); ++l)
                                if (torsion.at(divisorIdx.at(l)).mod
                                        == newPrimePower) {
                                    newIdx = l;
                                    break;
                                }
                            if (newIdx == divisorIdx.size())
                                for (unsigned l = 0; l < torsion.size(); ++l)
                                    if (torsion.at(l).mod == newPrimePower) {
                                        divisorIdx.push_back(l);
                                        break;
                                    }
                            if (newIdx == divisorIdx.size()) {
                                torstruct newTS;
                                newTS.mod = newPrimePower;
                                newTS.dim.resize(ccSpacesSize);
                                for (std::vector<std::vector<
                                        unsigned> >:: iterator vvit =
                                        newTS.dim.begin();
                                        vvit != newTS.dim.end(); ++vvit)
                                    vvit->resize(QIdx);
                                torsion.push_back(newTS);
                                divisorIdx.push_back(torsion.size() - 1);
                            }
                            torsion.at(divisorIdx.at(newIdx)).dim.at(
                                    i + 1).at(j) += 1;
                        }
                    }
                    prevF = f;
                }
                ranks.at(i).at(j) = r;
            }
        }
    }
    pari_close();
}
